from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models


class PFPContentType(models.Model):
    """ContentTypes to create PerFieldPermissions for."""
    content_type = models.OneToOneField(ContentType, unique=True, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.content_type)


class PFPRoleType(models.Model):
    """Role types (e.g. User, Group) to allow users to filter available roles with."""
    role_ctype = models.OneToOneField(ContentType, unique=True, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.role_ctype)


class PFPFilterType(models.Model):
    """ContentTypes to make available when managing permissions."""
    filter_ctype = models.OneToOneField(ContentType, unique=True, on_delete=models.CASCADE)

    def __str__(self):
        return str(self.role_ctype)


class PerFieldPermission(Permission):
    """Extends Permission to enable field level permissions."""
    field_name = models.CharField(max_length=100, blank=False, null=False)
    model_permission = models.ForeignKey(
            Permission,
            null=False,
            blank=False,
            on_delete=models.CASCADE,
            related_name='field_permissions',
            )

    class Meta:
        permissions = (('manage_permissions', 'Can manage permissions'),)

    def clean(self):
        """Ensure a PFP has the same ContentType as its parent Permission."""
        if self.content_type_id != self.model_permission.content_type_id:
            raise ValidationError('PerFieldPermission model_permission must be of the same '
                                  'ContentType as self.')
        super().clean()

    def validate_unique(self, exclude=None):
        """
        Ensure only one PFP exists for a given Permission/field combination.
        """
        super().validate_unique(exclude)
        try:
            perm = PerFieldPermission.objects.get(
                    codename=self.codename,
                    model_permission=self.model_permission,
                    )
        except ObjectDoesNotExist:
            return

        if ((hasattr(self, 'pk') and self.pk != perm.pk)
                or not hasattr(self, 'pk')):
            raise ValidationError("Only one PerFieldPermission allowed for each model "
                                  "permission/field combination.")



from functools import update_wrapper

from django.conf.urls import url
from django.contrib import admin
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied, ValidationError
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db import transaction
from django.db.models import Q
from django.template.response import TemplateResponse

from perfieldperms import forms
from perfieldperms import models
from perfieldperms.utils import get_m2m_remote_fields, get_manager_name, get_unpermitted_fields


class PFPModelAdmin(admin.ModelAdmin):
    """
    Extends get_form() to disable/remove fields based on user's field level
    permissions.

    Attributes
    ----------
    pfp_disable_fields : bool
                         Whether to disable fields or remove them from the ModelForm when applying
                         per-field permissions
    """
    pfp_disable_fields = True

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.disabled_fields = get_unpermitted_fields(self.model, request.user, obj)
        form.disable_fields = self.pfp_disable_fields
        return form


class PFPInlineAdmin(admin.options.InlineModelAdmin):
    """
    Extends get_formset() to disable fields based on user's field level
    permissions.

    Attributes
    ----------
    pfp_disable_fields : bool
                         Whether to disable fields or remove them from the ModelForm when applying
                         per-field permissions
    """
    pfp_disable_fields = True

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj, **kwargs)
        form = formset.form
        form.disabled_fields = get_unpermitted_fields(self.model, request.user, obj)
        form.disable_fields = self.pfp_disable_fields
        return formset


class PerFieldPermissionAdmin(admin.ModelAdmin):
    """
    Extends ModelAdmin by adding a /manage/ url which provides an alternative
    table layout for managing permissions.
    """
    def _get_filter_form_initial(self, request):
        """
        Extracts fields needed to initialise the filter forms from submitted
        data. Uses said forms to verify the submitted data.

        Returns a dict of values needed to initialise the filter forms.
        """
        required_fields = [
                'pfp_pfilter-model', 'pfp_pfilter-show_pfps',
                'pfp_rfilter-role_type',
                ]
        initial = {}
        if request.method == 'GET':
            init_data = self.get_changeform_initial_data(request)
        elif request.method == 'POST':
            init_data = request.POST.copy()
        # Stick the GET or POST data into the form to validate it, and if it
        # validates grab the cleaned data from the required fields which we'll
        # use to populate the non-required fields later on
        init_data = {k: v for k, v in init_data.items() if k in required_fields}
        init_form = forms.PFPPermFilterForm(data=init_data)
        if init_form.is_bound and init_form.is_valid():
            initial['model'] = init_form.cleaned_data['model']
            initial['show_pfps'] = init_form.cleaned_data['show_pfps']

        init_form = forms.PFPRoleFilterForm(data=init_data)
        if init_form.is_bound and init_form.is_valid():
            initial['role_type'] = init_form.cleaned_data['role_type']

        return initial

    def _get_perm_choices(self, cleaned_data):
        """
        Constructs the list of permissions that will be offered to each
        role in the selection matrix. Expects cleaned data from a
        PFPManagePermsForm.

        Returns a Permission queryset.
        """
        model_ctype = cleaned_data['model']
        perms = [perm.id for perm in cleaned_data['permissions']]

        if cleaned_data['model_fields']:
            fields = cleaned_data['model_fields']

            # If fields and permissions have been selected, then for any model
            # level permissions selected get their linked pfps, and include any
            # directly selected pfps. Otherwise, just get pfps for the
            # fields selected.
            if perms:
                model_perms = Permission.objects.filter(
                        perfieldpermission__isnull=True,
                        id__in=perms,
                        ).values_list('id', flat=True)
                pfp_ids = set(perms).difference(set(model_perms))
                perms_qs = Permission.objects.filter(
                        Q(perfieldpermission__field_name__in=fields,
                            perfieldpermission__model_permission_id__in=model_perms)
                        | Q(id__in=pfp_ids)
                        ).distinct().order_by()
            else:
                perms_qs = Permission.objects.filter(
                        content_type=model_ctype,
                        perfieldpermission__field_name__in=fields
                        ).order_by()
        # If fields aren't set then we use the selected permissions, or all
        # available permissions
        elif cleaned_data['permissions']:
            perms_qs = cleaned_data['permissions']
        elif cleaned_data['show_pfps']:
            perms_qs = Permission.objects.filter(
                    content_type=model_ctype).order_by()
        else:
            perms_qs = Permission.objects.filter(
                    content_type=model_ctype,
                    perfieldpermission__isnull=True,
                    ).order_by()
        return perms_qs

    def _get_perm_select_initial(self, request, clean_data, perms_qs):
        """
        Get info about which permissions roles currently possess so we can
        intialise the selection form. Kind of expensive in terms of queries but
        we shouldn't be doing this often.

        Should split this into getting roles, and creating initial tuple.

        Returns a tuple to pass to the formset generated by
        pfp_perm_select_form_factory(), and a Page object for paginating the
        display of the selected roles.
        """
        role_type = clean_data['role_type']
        role_class = role_type.model_class()
        user_model = get_user_model()
        remote_fields = get_m2m_remote_fields(Permission, role_class)
        if len(remote_fields) != 1:
            raise ValidationError('The role type selected has no, or multiple relations to the '
                                  'Permissions model. Expected to find one.')
        else:
            role_related_query = remote_fields[0].name

        initial = []
        all_roles = role_class.objects.all().order_by('pk')

        # Paginate list of roles
        paginator = Paginator(all_roles, 25)
        page = request.GET.get('page')
        try:
            roles = paginator.page(page)
        except PageNotAnInteger:
            roles = paginator.page(1)
        except EmptyPage:
            roles = paginator.page(paginator.num_pages)

        for role in roles:
            perms = perms_qs.filter(**{role_related_query: role}).order_by()
            if role_class == user_model:
                if role.get_full_name():
                    role_name = role.get_full_name()
                else:
                    role_name = role.username
            else:
                role_name = role.name
            initial.append({
                    'role_name': role_name,
                    'role_id': str(role.id),
                    'permissions': perms,
                    })
        return initial, roles

    def _process_perm_forms(self, request, clean_data, perm_formset):
        """
        Add and remove permissions based on form submission.

        Since it's possible users have been added/removed since the list of users was loaded, some
        forms in the formset may have IDs that don't match with what was loaded when the user last
        accessed the form. Skip the user as it would be a pain to try and match them up to an entry
        in the new list of users.
        """
        if perm_formset.forms:
            # perms_offered = perm_formset.forms[0].fields['permissions'].queryset
            role_class = clean_data['role_type'].model_class()
            manager_name = get_manager_name(role_class, Permission)

        for form in perm_formset.forms:
            if form.has_changed() and 'permissions' in form.changed_data:
                if 'role_id' in form.changed_data:
                    messages.warning(
                            request,
                            'Permissions for {} were not updated due to database changes since '
                            'loading the form. Please try again.'.format(
                                form.cleaned_data['role_name']
                                ),
                            )
                    continue

                role_id = form.cleaned_data['role_id']
                selected_perms = set(form.cleaned_data['permissions'])
                try:
                    role = role_class.objects.get(id=role_id)
                except ObjectDoesNotExist:
                    messages.error(
                            request,
                            'Permissions for {} were not updated because the role could not be '
                            'found.'.format(form.cleaned_data['role_name']),
                            )
                    continue

                # Get perms to add/remove
                init_perms = set(form.initial['permissions'])
                del_perms = init_perms.difference(
                        form.cleaned_data['permissions'])
                add_perms = selected_perms.difference(init_perms)

                # Process
                perm_manager = getattr(role, manager_name)
                with transaction.atomic():
                    perm_manager.remove(*del_perms)
                    perm_manager.add(*add_perms)

                messages.success(
                        request,
                        'Permissions for {} were changed. Added: {}. Removed: {}.'.format(
                            form.cleaned_data['role_name'],
                            len(add_perms),
                            len(del_perms),
                            )
                        )

    def manage_view(self, request):
        """
        Provide a filterable grid of roles/permissions as an alternative to
        editing each user/role individually.
        """
        if not(request.user.has_perm('perfieldperms.manage_permissions')):
            raise PermissionDenied

        context = {}
        template = '{}/perfieldperm_manage.html'.format(
                self.model._meta.app_label
                )
        page = None
        initial = self._get_filter_form_initial(request)
        data = request.POST
        perm_filter_form = forms.PFPPermFilterForm(initial=initial, data=data)
        role_filter_form = forms.PFPRoleFilterForm(initial=initial, data=data)

        if ((role_filter_form.is_bound and role_filter_form.is_valid())
                and
                (perm_filter_form.is_bound and perm_filter_form.is_valid())):
            # Filter forms are valid so see if we have a valid selection form
            clean_data = perm_filter_form.cleaned_data.copy()
            clean_data.update(role_filter_form.cleaned_data)
            perms_qs = self._get_perm_choices(clean_data)
            PermSelectFormSet = forms.pfp_perm_select_form_factory(perms_qs)
            # Generate initial data for formset which will be a list of roles
            # and any permissions they currently have.
            perm_select_initial, page = self._get_perm_select_initial(
                    request,
                    clean_data,
                    perms_qs,
                    )

            if '_save' in request.POST or '_continue' in request.POST:
                perm_formset = PermSelectFormSet(
                        initial=perm_select_initial,
                        data=data,
                        )
                try:
                    if perm_formset.is_bound and perm_formset.is_valid():
                        self._process_perm_forms(request, clean_data, perm_formset)
                except ValidationError:
                    messages.error(request, 'The form failed to validate, probably due to changes '
                                            'to filter settings.')
            else:
                perm_formset = PermSelectFormSet(initial=perm_select_initial)
            context['perm_formset'] = perm_formset

        # Update context with info required by the AdminSite, and the template
        context.update(self.admin_site.each_context(request))
        context.update(dict(
                opts=self.model._meta,
                title='Manage permissions',
                save_as=self.save_as,
                save_as_continue=self.save_as_continue,
                show_save=True,
                show_save_and_continue=True,
                show_save_and_add_another=False,
                show_save_as_new=False,
                perm_filter_form=perm_filter_form,
                role_filter_form=role_filter_form,
                page=page,
                ))
        return TemplateResponse(request, template, context=context)

    def get_urls(self):
        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        urls = [
                url(r'^manage/$', wrap(self.manage_view),
                    name='{0[0]}_{0[1]}_manage'.format(info)),
                ]
        return urls + super().get_urls()


admin.site.register(models.PerFieldPermission, PerFieldPermissionAdmin)


@admin.register(models.PFPContentType)
class PFPContentTypeAdmin(admin.ModelAdmin):
    form = forms.PFPContentTypesForm


@admin.register(models.PFPRoleType)
class PFPRoleTypeAdmin(admin.ModelAdmin):
    form = forms.PFPRoleTypesForm



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import Permission
from django.core.exceptions import ObjectDoesNotExist

from perfieldperms.models import PerFieldPermission
from perfieldperms.utils import cache_to_set, make_model_perm_sets, make_pfp_perm_sets


class PFPBackend(object):
    """
    Provide permission checking hooks for PerFieldPermissions.
    """
    mbackend = ModelBackend()

    def authenticate(self, request, username=None, password=None, **kwargs):
        """Use the Django default backend for authentication."""
        return self.mbackend.authenticate(request, username=username, password=password, **kwargs)

    @staticmethod
    def _get_user_permissions(user_obj):
        # Get standard perms
        perms = user_obj.user_permissions.filter(
                perfieldpermission__isnull=True,
                ).values_list('content_type__app_label', 'codename')
        perm_sets = make_model_perm_sets(perms)
        # Get per field perms
        pfps = PerFieldPermission.objects.filter(
                user=user_obj
                ).values_list(
                'content_type__app_label',
                'codename',
                'model_permission__codename',
                )
        perm_sets = make_pfp_perm_sets(pfps, perm_sets)
        return perm_sets

    @staticmethod
    def _get_group_permissions(user_obj):
        user_groups_field = get_user_model()._meta.get_field('groups')
        user_groups_query = 'group__{}'.format(
                user_groups_field.related_query_name()
                )
        perms = Permission.objects.filter(**{
                'perfieldpermission__isnull': True,
                user_groups_query: user_obj
                }).values_list('content_type__app_label', 'codename')
        perm_sets = make_model_perm_sets(perms)
        pfps = PerFieldPermission.objects.filter(
                **{user_groups_query: user_obj}
                ).values_list(
                'content_type__app_label',
                'codename',
                'model_permission__codename',
                )
        perm_sets = make_pfp_perm_sets(pfps, perm_sets)
        return perm_sets

    def _get_permissions(self, user_obj, obj, from_name):
        """
        Lookup group or user permissions for `user_obj` depending on the
        contents of `from_name`. `from_name` can ber "user" or "group", and
        calls `_get_user_permissions` and `_get_group_permissions`
        respectively.
        """
        if not user_obj.is_active or user_obj.is_anonymous:
            return set()

        cache_name = '_pfp_{}_perm_cache'.format(from_name)
        if not hasattr(user_obj, cache_name):
            if user_obj.is_superuser:
                perms = Permission.objects.filter(
                        perfieldpermission__isnull=True,
                        ).values_list('content_type__app_label', 'codename')
                perm_sets = make_model_perm_sets(perms)
            else:
                perm_sets = getattr(self, '_get_{}_permissions'.format(from_name))(user_obj)
            setattr(user_obj, cache_name, perm_sets)
        return cache_to_set(getattr(user_obj, cache_name))

    def get_user_permissions(self, user_obj, obj=None):
        """Get permissions granted only to this user."""
        return self._get_permissions(user_obj, obj, 'user')

    def get_group_permissions(self, user_obj, obj=None):
        """Get permissions granted to groups this user is a member of."""
        return self._get_permissions(user_obj, obj, 'group')

    def get_all_permissions(self, user_obj, obj=None):
        """
        Get all user and role based permissions for this user. Creates a cache of permissions for
        life of user object.
        """
        if not user_obj.is_active or user_obj.is_anonymous:
            return set()

        if not hasattr(user_obj, '_pfp_perm_cache'):
            self.get_user_permissions(user_obj, obj)
            self.get_group_permissions(user_obj, obj)
            # Copy the group perm cache, then update it from the user cache
            # effectively overriding those settings
            user_obj._pfp_perm_cache = user_obj._pfp_group_perm_cache.copy()
            user_obj._pfp_perm_cache.update(user_obj._pfp_user_perm_cache)
            # The below will merge caches instead of overwriting
            # user_obj._pfp_perm_cache = self._merge_caches(
            #        user_obj._pfp_perm_cache,
            #        user_obj._pfp_user_perm_cache,
            #        )
        return cache_to_set(user_obj._pfp_perm_cache)

    def has_perm(self, user_obj, perm, obj=None):
        """
        Returns true if `user_obj` has `perm` in their set of all permissions.
        """
        if not user_obj.is_active:
            return False
        if not hasattr(user_obj, '_pfp_perm_cache'):
            if not self.get_all_permissions(user_obj, obj):
                return False
        cache = user_obj._pfp_perm_cache
        # Looking for model level perm, no need to go deeper
        if perm in cache:
            return True
        perm_split = perm.rsplit('__', maxsplit=1)
        perm_key = perm_split[0]
        # A user has a field level permission if the model level parent is set
        # and no fields are set, or the field permission is set.
        return perm_key in cache and (
                not cache[perm_key] or perm in cache[perm_key]
                )

    def has_module_perms(self, user_obj, app_label):
        """
        Returns True if `user_obj` has any permissions in the given
        `app_label`.
        """
        if not user_obj.is_active:
            return False
        if not hasattr(user_obj, '_pfp_perm_cache'):
            if not self.get_all_permissions(user_obj):
                return False
        for perm in user_obj._pfp_perm_cache.keys():
            if perm[:perm.index('.')] == app_label:
                return True
        return False

    def has_all_fields(self, user_obj, perm, obj=None):
        """
        Return if `user_obj` has all field level permissions for a given
        permission, in which case the set of field permissions will be empty
        """
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return False

        if isinstance(perm, type(str())):
            app_label, codename = perm.split('.')
            try:
                perm = PerFieldPermission.objects.get(
                    content_type__app_label=app_label,
                    codename=codename,
                    )
            except ObjectDoesNotExist:
                perm = Permission.objects.get(
                    content_type__app_label=app_label,
                    codename=codename,
                    )
        if isinstance(perm, PerFieldPermission):
            perm_key = '{}.{}'.format(
                    perm.content_type.app_label,
                    perm.model_permission.codename,
                    )
        elif isinstance(perm, Permission):
            perm_key = '{}.{}'.format(perm.content_type.app_label, perm.codename)
        else:
            raise TypeError("""Perm must be a string representation of a
            Permission, a Permission, or a PerFieldPermission""")

        if not hasattr(user_obj, '_pfp_perm_cache'):
            self.get_all_permissions(user_obj, obj)
        cache = user_obj._pfp_perm_cache
        return perm_key in cache and not cache[perm_key]

    def get_user(self, user_id):
        return self.mbackend.get_user(user_id)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-02-21 05:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='PerFieldPermission',
            fields=[
                ('permission_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='auth.Permission')),
                ('field_name', models.CharField(max_length=100)),
                ('model_permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='field_permissions', to='auth.Permission')),
            ],
            options={
                'permissions': (('manage_permissions', 'Can manage permissions'),),
            },
            bases=('auth.permission',),
        ),
        migrations.CreateModel(
            name='PFPContentType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content_type', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='PFPFilterType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('filter_ctype', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='PFPRoleType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role_ctype', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
    ]



# Generated by Django 2.1.1 on 2018-09-21 06:11

import django.contrib.auth.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('perfieldperms', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='perfieldpermission',
            managers=[
                ('objects', django.contrib.auth.models.PermissionManager()),
            ],
        ),
    ]



from django import template

register = template.Library()


@register.inclusion_tag('perfieldperms/pfp_submit_line.html', takes_context=True)
def pfp_submit_row(context):
    ctx = context.update(dict(
            show_save=True,
            show_delete_link=False,
            show_save_as_new=False,
            show_save_and_add_another=False,
            show_save_and_continue=True,
            ))
    return ctx


@register.inclusion_tag('perfieldperms/pfp_filter_field.html')
def pfp_filter_field(field):
    ctx = {'field': field}
    return ctx



import os
from setuptools import find_packages, setup


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__),
    os.pardir)))
    
setup(
        name='django-perfieldperms',
        version='0.1.6',
        packages=find_packages(exclude=['tests*']),
        include_package_data=True,
        license='Apache-2.0',
        description='Per model field permissions for Django.',
        url='https://gitlab.anu.edu.au/CIMS/django-perfieldperms',
        author='Lincoln Smith',
        author_email='lincoln.smith@anu.edu.au',
        classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3 :: Only',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Development Status :: 3 - Alpha',
            'Environment :: Web Environment',
            'Framework :: Django :: 1.11',
            'Framework :: Django :: 2.0',
            'Framework :: Django :: 2.1',
            'Intended Audience :: Developers',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Software Development :: Libraries :: Python Modules',
            ],
        install_requires=['django>=1.11'],
        python_requires='~=3.4',
        long_description="""\
Implements model field permissions for Django.

Provides models and an authentication/authorisation backend to enable permissions to be set at the
field level of a django model. Also provides Django admin and form classes to assist with
integration of field level permissions into the rest of Django.
""",
        )

